#![no_main]
#![no_std]

extern crate cortex_m;
extern crate cortex_m_rt as rt;
//extern crate panic_reset;
extern crate panic_semihosting;
extern crate stm32f4xx_hal as hal;

mod packet;
mod spi_data;

#[rtic::app(device = stm32f4xx_hal::pac, peripherals = true, dispatchers = [USART6])]
mod app {
    use hal::dma;
    use hal::gpio::{self, Speed};
    use hal::pac::{self, SPI1, TIM5};
    use hal::prelude::*;
    use hal::spi;
    use hal::spi::{Mode, Phase, Polarity, Spi};
    use hal::timer::Delay;
    use rf4463::config::RADIO_CONFIG_500_2;
    use rf4463::Rf4463;
    use ringbuffer::{ConstGenericRingBuffer, RingBuffer, RingBufferRead, RingBufferWrite};
    use systick_monotonic::*;

    use crate::packet::Packet;
    use crate::packet::PACKET_LEN;
    use crate::spi_data::{self, SpiData};

    // in # packets
    const BUFFER_LEN: usize = 100;

    // in bytes
    // explicitly make this smaller than a packet so that we don't buffer things in the tx DMA for a while
    const PI_RX_BUFFER_LEN: usize = 256;

    // in bytes
    const PI_TX_BUFFER_LEN: usize = spi_data::LEN;

    const SYS_CLK: u32 = 100_000_000;

    const MODE: Mode = Mode {
        polarity: Polarity::IdleLow,
        phase: Phase::CaptureOnFirstTransition,
    };

    type Radio = Rf4463<
        PACKET_LEN,
        Spi<SPI1>,
        gpio::Pin<'B', 1, gpio::Output>,
        gpio::Pin<'B', 0, gpio::Output>,
        Delay<TIM5, 1000000>,
    >;

    type RxTransfer = dma::Transfer<
        dma::StreamX<pac::DMA1, 3>,
        0,
        spi::Rx<pac::SPI2>,
        dma::PeripheralToMemory,
        &'static mut [u8; PI_RX_BUFFER_LEN],
    >;

    type TxTransfer = dma::Transfer<
        dma::StreamX<pac::DMA1, 4>,
        0,
        spi::Tx<pac::SPI2>,
        dma::MemoryToPeripheral,
        &'static mut [u8; PI_TX_BUFFER_LEN],
    >;

    #[derive(Debug)]
    enum SlaveCmd {
        SendPacket,
    }

    impl SlaveCmd {
        pub fn from_u8(i: u8) -> Option<Self> {
            match i {
                0x01 => Some(Self::SendPacket),
                _ => None,
            }
        }
    }

    #[allow(clippy::large_enum_variant)]
    pub enum SlaveState {
        Idle,
        RecvPacket(Packet),
    }

    #[shared]
    struct Shared {
        radio: Radio,
        tx_buf: ConstGenericRingBuffer<Packet, BUFFER_LEN>,

        pi_tx: TxTransfer,
        other_tx_buf: Option<&'static mut [u8; PI_TX_BUFFER_LEN]>,
        spi_data: SpiData,
    }

    #[local]
    struct Local {
        radio_irq: gpio::Pin<'B', 2, gpio::Input>,
        state: SlaveState,

        pi_rx: RxTransfer,
        other_rx_buf: Option<&'static mut [u8; PI_RX_BUFFER_LEN]>,
    }

    #[monotonic(binds = SysTick, default = true)]
    type Tonic = Systick<10000>;

    #[init]
    fn init(mut ctx: init::Context) -> (Shared, Local, init::Monotonics) {
        let rcc = ctx.device.RCC.constrain();

        let gpioa = ctx.device.GPIOA.split();
        let gpiob = ctx.device.GPIOB.split();

        // setup clocks
        let clocks = rcc
            .cfgr
            .use_hse(25.MHz())
            .sysclk(SYS_CLK.Hz())
            .pclk1(48.MHz())
            .pclk2(48.MHz())
            .freeze();
        let mut sys_cfg = ctx.device.SYSCFG.constrain();

        // setup our fake reset pin (temporary)
        let mut reset_pin = gpiob.pb12.into_pull_up_input();
        reset_pin.make_interrupt_source(&mut sys_cfg);
        reset_pin.enable_interrupt(&mut ctx.device.EXTI);
        reset_pin.trigger_on_edge(&mut ctx.device.EXTI, gpio::Edge::Rising);

        // setup 4463 spi
        let mosi = gpioa.pa7.into_alternate().speed(Speed::VeryHigh);
        let miso = gpioa.pa6.into_alternate().speed(Speed::VeryHigh);
        let sclk = gpioa.pa5.into_alternate().speed(Speed::VeryHigh);

        let sdn = gpiob.pb1.into_push_pull_output();
        let cs = gpiob.pb0.into_push_pull_output();
        let mut radio_irq = gpiob.pb2.into_pull_up_input();

        let spi = Spi::new(ctx.device.SPI1, (sclk, miso, mosi), MODE, 10.MHz(), &clocks);

        let delay = ctx.device.TIM5.delay_us(&clocks);
        let radio = Rf4463::new(spi, sdn, cs, delay, &mut RADIO_CONFIG_500_2.clone()).unwrap();

        radio_irq.make_interrupt_source(&mut sys_cfg);
        radio_irq.enable_interrupt(&mut ctx.device.EXTI);
        radio_irq.trigger_on_edge(&mut ctx.device.EXTI, gpio::Edge::Falling);

        // setup Pi SPI
        let pi_mosi = gpiob.pb15.internal_resistor(gpio::Pull::Up);
        let pi_miso = gpiob.pb14.internal_resistor(gpio::Pull::Down);
        let pi_sclk = gpiob.pb13.internal_resistor(gpio::Pull::Down);

        // TODO nss?

        let mut pi_spi = ctx
            .device
            .SPI2
            .spi_slave((pi_sclk, pi_miso, pi_mosi, None), MODE);
        pi_spi.set_internal_nss(false);

        let (tx, rx) = pi_spi.use_dma().txrx();

        let streams = dma::StreamsTuple::new(ctx.device.DMA1);

        let spi_data = SpiData::new(0.0, true);
        let tx_buffer = cortex_m::singleton!(: [u8; PI_TX_BUFFER_LEN] = spi_data.into()).unwrap();
        let other_tx_buf =
            Some(cortex_m::singleton!(: [u8; PI_TX_BUFFER_LEN] = spi_data.into()).unwrap());

        let mut pi_tx = dma::Transfer::init_memory_to_peripheral(
            streams.4,
            tx,
            tx_buffer,
            None,
            dma::config::DmaConfig::default()
                .memory_increment(true)
                .fifo_enable(true)
                .fifo_error_interrupt(true),
        );

        // Enable circular mode
        // safety: probably?
        unsafe {
            let dma1 = &*pac::DMA1::ptr();
            dma1.st[4].cr.modify(|_, w| w.circ().set_bit());
        }

        pi_tx.start(|_tx| {});

        let rx_buffer =
            cortex_m::singleton!(: [u8; PI_RX_BUFFER_LEN] = [0u8; PI_RX_BUFFER_LEN]).unwrap();
        let other_rx_buf =
            Some(cortex_m::singleton!(: [u8; PI_RX_BUFFER_LEN] = [0u8; PI_RX_BUFFER_LEN]).unwrap());

        let mut pi_rx = dma::Transfer::init_peripheral_to_memory(
            streams.3,
            rx,
            rx_buffer,
            None,
            dma::config::DmaConfig::default()
                .memory_increment(true)
                .fifo_enable(true)
                .fifo_error_interrupt(true)
                .transfer_complete_interrupt(true),
        );

        pi_rx.start(|_rx| {});

        // start our radio temp loop
        get_radio_temp::spawn_after(1u64.secs()).unwrap();

        // start our radio interrupt loop
        // TODO I kind of hate this. We need this because sometimes we miss the real interrupt
        // I don't know why
        run_radio_interrupt::spawn_after(500u64.micros()).unwrap();

        (
            Shared {
                radio,
                tx_buf: ConstGenericRingBuffer::new(),

                pi_tx,
                other_tx_buf,

                spi_data,
            },
            Local {
                radio_irq,
                state: SlaveState::Idle,

                pi_rx,
                other_rx_buf,
            },
            init::Monotonics(Systick::new(ctx.core.SYST, SYS_CLK)),
        )
    }

    #[idle(shared=[radio, tx_buf, pi_tx, other_tx_buf, spi_data])]
    fn idle(mut ctx: idle::Context) -> ! {
        let mut iterations_since_last_packet = 0;
        loop {
            if let Some(pkt) = ctx.shared.tx_buf.lock(|buf| buf.dequeue()) {
                if iterations_since_last_packet >= 1000 {
                    // if we haven't transmitted in a while, the receiver's AGC is going to be messed up
                    // send 5 packets just so it can calibrate before we actually send it
                    for _ in 0..5 {
                        const GARBAGE: [u8; PACKET_LEN] = [0x00; PACKET_LEN];
                        ctx.shared.radio.lock(|r| r.start_tx(GARBAGE).unwrap());

                        // wait for radio to be idle again
                        loop {
                            if ctx.shared.radio.lock(|r| r.is_idle()) {
                                break;
                            }
                        }
                    }
                }

                ctx.shared.radio.lock(|r| r.start_tx(pkt.data).unwrap());

                // wait for radio to be idle again
                loop {
                    if ctx.shared.radio.lock(|r| r.is_idle()) {
                        break;
                    }
                }

                // update our SPI bus to say if we have free space or not
                let free = ctx.shared.tx_buf.lock(|b| b.len() < b.capacity() - 2);
                let mut new_spi_data = ctx.shared.spi_data.lock(|sd| *sd);
                new_spi_data.set_free(free);

                set_spi_tx(
                    &mut ctx.shared.pi_tx,
                    &mut ctx.shared.other_tx_buf,
                    &mut ctx.shared.spi_data,
                    new_spi_data,
                );

                iterations_since_last_packet = 0;
            } else {
                iterations_since_last_packet += 1;
            }
        }
    }

    #[task(priority = 1, shared = [radio, spi_data, pi_tx, other_tx_buf])]
    fn get_radio_temp(mut ctx: get_radio_temp::Context) {
        // get temp every second

        if let Ok(new_temp) = ctx.shared.radio.lock(|r| r.get_temp()) {
            let mut new_spi_data = ctx.shared.spi_data.lock(|sd| *sd);
            new_spi_data.set_temp(new_temp);

            set_spi_tx(
                &mut ctx.shared.pi_tx,
                &mut ctx.shared.other_tx_buf,
                &mut ctx.shared.spi_data,
                new_spi_data,
            );
        }

        get_radio_temp::spawn_after(1u64.secs()).unwrap();
    }

    #[task(priority = 1, shared = [radio])]
    fn run_radio_interrupt(mut ctx: run_radio_interrupt::Context) {
        ctx.shared.radio.lock(|r| r.interrupt().unwrap());

        run_radio_interrupt::spawn_after(500u64.micros()).unwrap();
    }

    #[task(binds = DMA1_STREAM3, priority = 2, local = [state, pi_rx, other_rx_buf], shared = [tx_buf, pi_tx, other_tx_buf, spi_data])]
    fn pi_spi_recv(mut ctx: pi_spi_recv::Context) {
        let rx = ctx.local.pi_rx;

        let (old_buf, _) = rx
            .next_transfer(ctx.local.other_rx_buf.take().unwrap())
            .unwrap();

        for b in old_buf.iter() {
            handle_incoming_byte(
                *b,
                ctx.local.state,
                &mut ctx.shared.tx_buf,
                &mut ctx.shared.pi_tx,
                &mut ctx.shared.other_tx_buf,
                &mut ctx.shared.spi_data,
            );
        }

        *ctx.local.other_rx_buf = Some(old_buf);

        rx.clear_transfer_complete_interrupt();
        rx.clear_fifo_error_interrupt();
    }

    #[task(binds = DMA1_STREAM4, priority = 2, shared = [pi_tx])]
    fn pi_spi_tx(mut ctx: pi_spi_tx::Context) {
        ctx.shared.pi_tx.lock(|tx| tx.clear_fifo_error_interrupt());
    }

    #[task(binds = EXTI2, shared = [radio], local = [radio_irq])]
    fn radio_irq(mut ctx: radio_irq::Context) {
        ctx.shared.radio.lock(|r| r.interrupt()).unwrap();

        let irq = ctx.local.radio_irq;
        if irq.is_high() {
            irq.clear_interrupt_pending_bit();
        }
    }

    // temporary. Eventually we'll want the pi to be connected directly to the reset pin
    #[task(binds = EXTI15_10, priority = 5)]
    fn reset(_: reset::Context) {
        cortex_m::peripheral::SCB::sys_reset();
    }

    fn handle_incoming_byte<
        T: rtic::Mutex<T = ConstGenericRingBuffer<Packet, BUFFER_LEN>>,
        A: rtic::Mutex<T = TxTransfer>,
        B: rtic::Mutex<T = Option<&'static mut [u8; PI_TX_BUFFER_LEN]>>,
        C: rtic::Mutex<T = SpiData>,
    >(
        x: u8,
        state: &mut SlaveState,
        buf: &mut T,
        tx: &mut A,
        other_tx: &mut B,
        spi_data: &mut C,
    ) {
        match state {
            SlaveState::Idle => {
                if let Some(cmd) = SlaveCmd::from_u8(x) {
                    match cmd {
                        SlaveCmd::SendPacket => {
                            let free = buf.lock(|b| b.len() < b.capacity() - 2);

                            let mut new_spi_data = spi_data.lock(|sd| *sd);
                            new_spi_data.set_free(free);

                            set_spi_tx(tx, other_tx, spi_data, new_spi_data);

                            *state = SlaveState::RecvPacket(Packet::default());
                        }
                    }
                }
            }

            SlaveState::RecvPacket(pkt) => {
                pkt.push(x);
                if pkt.is_complete() {
                    buf.lock(|b| b.enqueue(pkt.clone()));

                    *state = SlaveState::Idle;
                }
            }
        };
    }

    fn set_spi_tx<
        A: rtic::Mutex<T = TxTransfer>,
        B: rtic::Mutex<T = Option<&'static mut [u8; PI_TX_BUFFER_LEN]>>,
        C: rtic::Mutex<T = SpiData>,
    >(
        tx: &mut A,
        other_tx: &mut B,
        spi_data: &mut C,
        new_spi_data: SpiData,
    ) {
        let bytes: [u8; spi_data::LEN] = new_spi_data.into();

        // if the new data is equal to the old data, do nothing
        if spi_data.lock(|sd| <[u8; spi_data::LEN]>::from(*sd)) == bytes {
            return;
        }

        spi_data.lock(|sd| *sd = new_spi_data);

        other_tx.lock(|t| t.as_mut().unwrap().clone_from_slice(&bytes));

        other_tx.lock(|t| {
            let (tx, _) =
                tx.lock(|tx_transfer| tx_transfer.next_transfer(t.take().unwrap()).unwrap());

            t.replace(tx);
        });
    }
}
